# https://huggingface.co/learn/nlp-course/en/chapter7/6

# seed
import random 
import torch 
from transformers import set_seed, AutoTokenizer, DataCollatorForLanguageModeling, AutoConfig, get_scheduler
from tqdm.notebook import tqdm
import wandb
from torch.optim import AdamW
from accelerate import Accelerator
from torch.utils.data.dataloader import DataLoader
from datasets import Dataset, load_dataset, DatasetDict
from models.standard_gpt2 import *
from models.spec_gpt2 import *
from models.standard_llama2 import get_standard_llama2
from models.spec_llama2 import get_spec_llama2
from collections import defaultdict
import argparse
import shutil

os.environ["WANDB__SERVICE_WAIT"] = "3000"
import shlex
import sys
args_str = ' '.join(sys.argv[1:])  # Get the command string (excluding the script name)
args_list = shlex.split(args_str)  # Split the command string into individual arguments

parser = argparse.ArgumentParser(description='Specificity Language Modeling')
parser.add_argument('--net', default='gpt2') # gpt2 / gpt2-medium / gpt2-large / gpt2-xl / llama
parser.add_argument('--spec', default=False) # True / False
parser.add_argument('--dataset', default='wikitext') # codeparrot/ wikitext
parser.add_argument('--resume', default=False) # Continue from latest checkpoint
# args = parser.parse_args()
args, unknown = parser.parse_known_args(args_list)  # Parse the arguments
device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

# torch.manual_seed(42)
# random.seed(42)
# set_seed(42)

topk_value = 0.5
# topk = args.spec # which model to train (standard or topk)

checkpoints_base_path = 'checkpoints'
project_name = 'test'
access_token = "" # for llama


def save_checkpoint(model, optimizer, lr_scheduler, completed_steps, model_checkpoint_name):
    checkpoint_dir = f'{checkpoints_base_path}/{model_checkpoint_name}'
    os.makedirs(checkpoint_dir, exist_ok=True)
    latest_checkpoint_path = os.path.join(checkpoint_dir, 'latest_checkpoint.pth')
    previous_checkpoint_path = os.path.join(checkpoint_dir, 'previous_checkpoint.pth')

    # If a previous checkpoint exists, rename it to previous_checkpoint.pth
    if os.path.exists(latest_checkpoint_path):
        os.rename(latest_checkpoint_path, previous_checkpoint_path)

    # Save the current checkpoint as the latest checkpoint
    torch.save({
        'model_state_dict': model.state_dict(),
        'optimizer_state_dict': optimizer.state_dict(),
        'lr_scheduler_state_dict': lr_scheduler.state_dict(),
        'completed_steps': completed_steps,
        'wandb_id': wandb_id
    }, latest_checkpoint_path)

    # Delete any other checkpoints
    for file in os.listdir(checkpoint_dir):
        if file not in ['latest_checkpoint.pth', 'previous_checkpoint.pth']:
            os.remove(os.path.join(checkpoint_dir, file))


model_checkpoint_name = f'spec_{args.spec}_{args.net}_{args.dataset}'

def any_keyword_in_string(string, keywords):
    for keyword in keywords:
        if keyword in string:
            return True
    return False

split = "train"  # "valid"
filters = ["pandas", "sklearn", "matplotlib", "seaborn"]

if args.dataset == 'codeparrot':
    # ds_train = load_dataset("codeparrot/codeparrot-clean-train", split="train", cache_dir='/scratch/alpine/sash9910')
    ds_train = load_dataset("codeparrot/codeparrot-clean-train", split="train")
    # ds_valid = load_dataset("codeparrot/codeparrot-clean-valid", split="validation", cache_dir='/scratch/alpine/sash9910')
    ds_valid = load_dataset("codeparrot/codeparrot-clean-valid", split="validation")
elif args.dataset == 'wikitext':
    # dataset = load_dataset('Salesforce/wikitext', 'wikitext-103-raw-v1', cache_dir='/scratch/alpine/sash9910') # https://huggingface.co/datasets/Salesforce/wikitext
    dataset = load_dataset('Salesforce/wikitext', 'wikitext-103-raw-v1') # https://huggingface.co/datasets/Salesforce/wikitext
    ds_train = dataset['train']
    ds_train = ds_train.rename_column("text", "content") # rename so it will be in the same format as codeparrot
    ds_valid = dataset['validation']
    ds_valid = ds_valid.rename_column("text", "content") # rename so it will be in the same format as codeparrot
    

raw_datasets = DatasetDict(
    {
        "train": ds_train,  # .shuffle().select(range(50000)),
        "valid": ds_valid,  # .shuffle().select(range(500))
    }
)
context_length = 128

if 'gpt' in args.net:
    # tokenizer = AutoTokenizer.from_pretrained("gpt2")
    tokenizer = AutoTokenizer.from_pretrained("huggingface-course/code-search-net-tokenizer")
    # standard gpt2
    config = AutoConfig.from_pretrained(
        # "gpt2", # https://huggingface.co/transformers/v2.1.1/pretrained_models.html
        args.net,
        vocab_size=len(tokenizer),
        n_ctx=context_length,
        bos_token_id=tokenizer.bos_token_id,
        eos_token_id=tokenizer.eos_token_id, 
        torch_dtype=torch.float16
        # cache_dir='/scratch/alpine/sash9910'
        )
    batch_size = 32
    if args.spec == 'True':
        print('Training specificity model...')
        spec_gpt2 = Spec_GPT2LMHeadModel(config).to(device) # load specificity architecture
        spec_gpt2_model_size = sum(t.numel() for t in spec_gpt2.parameters() if t.requires_grad)
        print(f"Spec GPT-2 size: {spec_gpt2_model_size/1000**2:.1f}M parameters")        
        model = spec_gpt2

    elif args.spec == 'False': # standard
        print('Training standard model...')
        standard_gpt2 = standard_GPT2LMHeadModel(config).to(device)
        model_size = sum(t.numel() for t in standard_gpt2.parameters() if t.requires_grad)
        print(f"GPT-2 size: {model_size/1000**2:.1f}M parameters")
        model = standard_gpt2

    else:
        print('args parsing didnt succeed and everything is messed upppppp RUNNNNN')

elif 'llama' in args.net:
    tokenizer = AutoTokenizer.from_pretrained("meta-llama/Llama-2-7b-chat-hf", token=access_token)
    batch_size = 8
    if args.spec == 'True':
        model = get_spec_llama2()
    elif args.spec == 'False':
        model = get_standard_llama2()
    else:
        print('args parsing didnt succeed and everything is messed upppppp RUNNNNN')


num_params = sum(p.numel() for p in model.parameters())
print(f"Total number of parameters: {num_params}")
num_trainable_params = sum(p.numel() for p in model.parameters() if p.requires_grad)
print(f"Number of trainable parameters: {num_trainable_params}")


tokenizer.pad_token = tokenizer.eos_token
data_collator = DataCollatorForLanguageModeling(tokenizer, mlm=False)

def tokenize(element):
    outputs = tokenizer(
        element["content"],
        truncation=True,
        max_length=context_length,
        return_overflowing_tokens=True,
        return_length=True,
    )
    input_batch = []
    for length, input_ids in zip(outputs["length"], outputs["input_ids"]):
        if length == context_length:
            input_batch.append(input_ids)
    return {"input_ids": input_batch}


tokenized_dataset = raw_datasets.map(
    tokenize, batched=True, remove_columns=raw_datasets["train"].column_names
)

keytoken_ids = []
for keyword in [
    "plt",
    "pd",
    "sk",
    "fit",
    "predict",
    " plt",
    " pd",
    " sk",
    " fit",
    " predict",
    "testtest",
]:
    ids = tokenizer([keyword]).input_ids[0]
    if len(ids) == 1:
        keytoken_ids.append(ids[0])
    else:
        print(f"Keyword has not single token: {keyword}")

def keytoken_weighted_loss(inputs, logits, keytoken_ids, alpha=1.0):
    # Shift so that tokens < n predict n
    shift_labels = inputs[..., 1:].contiguous()
    shift_logits = logits[..., :-1, :].contiguous()
    # Calculate per-token loss
    loss_fct = CrossEntropyLoss(reduce=False)
    loss = loss_fct(shift_logits.view(-1, shift_logits.size(-1)), shift_labels.view(-1))
    # Resize and average loss per sample
    loss_per_sample = loss.view(shift_logits.size(0), shift_logits.size(1)).mean(axis=1)
    # Calculate and scale weighting
    # weights = torch.stack([(inputs == kt).float() for kt in keytoken_ids]).sum(
    #     axis=[0, 2])
    if not keytoken_ids:
        weights = torch.ones_like(inputs[:, 0])  # default value
    else:
        weights = torch.stack([(inputs == kt).float() for kt in keytoken_ids], dim=0).sum(axis=[0, 2])

    
    weights = alpha * (1.0 + weights)
    # Calculate weighted average
    weighted_loss = (loss_per_sample * weights).mean()
    return weighted_loss

tokenized_dataset.set_format("torch")
train_dataloader = DataLoader(tokenized_dataset["train"], batch_size=batch_size, shuffle=True)
eval_dataloader = DataLoader(tokenized_dataset["valid"], batch_size=batch_size) 

weight_decay = 0.1

def get_grouped_params(model, no_decay=["bias", "LayerNorm.weight"]):
    params_with_wd, params_without_wd = [], []
    for n, p in model.named_parameters():
        if any(nd in n for nd in no_decay):
            params_without_wd.append(p)
        else:
            params_with_wd.append(p)
    return [
        {"params": params_with_wd, "weight_decay": weight_decay},
        {"params": params_without_wd, "weight_decay": 0.0},
    ]

def evaluate():
    model.eval()
    losses = []
    for step, batch in enumerate(eval_dataloader):
        with torch.no_grad():
            outputs = model(batch["input_ids"], labels=batch["input_ids"])
            
        losses.append(outputs.loss)

    loss = torch.mean(torch.tensor(losses))
    try:
        perplexity = torch.exp(loss)
    except OverflowError:
        perplexity = float("inf")
    print(f'perplexity: {perplexity}')
    print(f'loss: {loss.item()}')
    return loss.item(), perplexity.item()    

optimizer = AdamW(get_grouped_params(model), lr=5e-4)
accelerator = Accelerator(mixed_precision="fp16")
model, optimizer, train_dataloader, eval_dataloader = accelerator.prepare(
    model, optimizer, train_dataloader, eval_dataloader
)

num_train_epochs = 30
num_update_steps_per_epoch = len(train_dataloader)
num_training_steps = num_train_epochs * num_update_steps_per_epoch

lr_scheduler = get_scheduler(
    name="linear",
    optimizer=optimizer,
    num_warmup_steps=1_000,
    num_training_steps=num_training_steps,
)

# Load checkpoint if continuing training
wandb_id = None

# Load checkpoint if continuing training
if args.resume == 'True':
    # Load latest checkpoint
    checkpoint_dir = f'{checkpoints_base_path}/{model_checkpoint_name}'
    latest_checkpoint_path = os.path.join(checkpoint_dir, 'latest_checkpoint.pth')
    previous_checkpoint_path = os.path.join(checkpoint_dir, 'previous_checkpoint.pth')

    checkpoint_path = None
    if os.path.exists(latest_checkpoint_path):
        try:
            checkpoint = torch.load(latest_checkpoint_path)
            checkpoint_path = latest_checkpoint_path
        except RuntimeError:
            pass
    if checkpoint_path is None and os.path.exists(previous_checkpoint_path):
        try:
            checkpoint = torch.load(previous_checkpoint_path)
            checkpoint_path = previous_checkpoint_path
        except RuntimeError:
            pass

    if checkpoint_path:
        model.load_state_dict(checkpoint['model_state_dict'])
        optimizer.load_state_dict(checkpoint['optimizer_state_dict'])
        lr_scheduler.load_state_dict(checkpoint['lr_scheduler_state_dict'])
        completed_steps = checkpoint['completed_steps']
        wandb_id = checkpoint['wandb_id']
        print(f'Loaded checkpoint from step {completed_steps}')
    else:
        wandb_id = None

    wandb.init(project=f"{project_name}", name=f'spec_{args.spec}_{args.net}_{args.dataset}', id=wandb_id, resume='allow' if wandb_id is not None else False)
else:
    wandb.init(project=f"{project_name}", name=f'spec_{args.spec}_{args.net}_{args.dataset}')
    wandb_id = wandb.run.id

gradient_accumulation_steps = 8
eval_steps = 1_00

model.train()
if checkpoint_path:
    completed_steps = checkpoint['completed_steps']
else:
    completed_steps = 0

step = completed_steps
for epoch in range(num_train_epochs):
    print(f'epoch: {epoch}')
    for batch_step, batch in tqdm(
        enumerate(train_dataloader, start=1), total=num_training_steps
    ):
        logits = model(batch["input_ids"]).logits
        # print('1')
        loss = keytoken_weighted_loss(batch["input_ids"], logits, keytoken_ids)
        # print('2')
        loss = loss / gradient_accumulation_steps
        # print('3')
        accelerator.backward(loss)
        # print('4')
        if batch_step % gradient_accumulation_steps == 0:
            accelerator.clip_grad_norm_(model.parameters(), 1.0)
            optimizer.step()
            lr_scheduler.step()
            optimizer.zero_grad()
            step += 1
        if (batch_step % (eval_steps * gradient_accumulation_steps)) == 0:
            eval_loss, perplexity = evaluate()
            accelerator.print({"loss/eval": eval_loss, "perplexity": perplexity})
            model.train()
            accelerator.wait_for_everyone()
            unwrapped_model = accelerator.unwrap_model(model)
            # Save checkpoint
            save_checkpoint(unwrapped_model, optimizer, lr_scheduler, step, model_checkpoint_name)

            # log stuff            
            wandb.log({"validation_loss": eval_loss}, step=step)
            wandb.log({"validation_perplexity": perplexity}, step=step)
            wandb.log({"train_loss": loss.item()}, step=step)
