# https://github.com/lucidrains/mlp-mixer-pytorch/blob/main/mlp_mixer_pytorch/mlp_mixer_pytorch.py
from torch import nn
from functools import partial
from einops.layers.torch import Rearrange, Reduce
import torch 
import spec_config
specificity = spec_config.spec_enabled
print(f'specificity: {specificity}')


pair = lambda x: x if isinstance(x, tuple) else (x, x)

class PreNormResidual(nn.Module):
    def __init__(self, dim, fn):
        super().__init__()
        self.fn = fn
        self.norm = nn.LayerNorm(dim)

    def forward(self, x):
        return self.fn(self.norm(x)) + x

def mask_3D_topk_row_wise(tensor, topk):
    '''
    my_tensor = 3D tensor of shape (batch_size, rows, cols) -- the input tensor
    top_value = percentage of range(0,1) -- what percentage of values to keep
    returns the masking tensor based on the input tensor
    masking row-wise
    '''    
    k = int(tensor.shape[-1] * topk)
    k = max(1, k)
    topgetValue, _ = tensor.topk(k, dim=-1)
    mask = tensor >= topgetValue[..., -1].unsqueeze(-1)
    return mask.float()    


# original (used to be called "FeedForward")
def OrigFeedForward(dim, expansion_factor = 4, dropout = 0., dense = nn.Linear):
    inner_dim = int(dim * expansion_factor)
    return nn.Sequential(
        dense(dim, inner_dim),
        nn.GELU(),
        nn.Dropout(dropout),
        dense(inner_dim, dim),
        nn.Dropout(dropout)
    )

if specificity == False:
    # original with tanh
    def FeedForward(dim, expansion_factor = 4, dropout = 0., dense = nn.Linear):
        inner_dim = int(dim * expansion_factor)
        return nn.Sequential(
            dense(dim, inner_dim),
            nn.Tanh(),
            nn.Dropout(dropout),
            dense(inner_dim, dim),
            nn.Dropout(dropout)
        )
else: 
    # specificity
    class FeedForward(nn.Module):
        def __init__(
                self, dim, expansion_factor = 4, dropout = 0., activation='tanh'):
            super().__init__()
            self.mlp_hidden_layer=1 # original has 2 layers
            # self.mlp_hidden_layer=3 # make it 4
            hidden_dim = int(dim * expansion_factor)

            if activation == 'gelu':
                self.act = nn.GELU()
            elif activation == 'softplus':
                self.act = nn.Softplus()
            elif activation == 'leaky':
                self.act = nn.LeakyReLU(0.1)    
            elif activation == 'tanh':
                self.act = nn.Tanh()  
            
            self.layers = nn.ModuleList()
            self.specs = nn.ModuleList()
            self.dropout = nn.Dropout(dropout)

            for i in range(self.mlp_hidden_layer + 1): 
                if i == 0:
                    self.layers.append(nn.Linear(dim, hidden_dim))
                    self.specs.append(nn.Linear(dim, hidden_dim, bias=False))
                elif i == self.mlp_hidden_layer:
                    self.layers.append(nn.Linear(hidden_dim, dim))
                else:
                    self.layers.append(nn.Linear(hidden_dim, hidden_dim))
                    self.specs.append(nn.Linear(hidden_dim, hidden_dim, bias=False))            

            # make sure the routing function is not trainable
            for spec_layer in self.specs:
                spec_layer.weight.requires_grad = False
                
            self.top_k = 0.5 # topk_rate 

        def forward(self, x):
            x_specs = x
            for i, layer in enumerate(self.layers[:-1]): 
                x_nn = self.act(layer(x))
                with torch.no_grad():
                    x_specs = self.specs[i](x_specs)
                    x_specs_mask = mask_3D_topk_row_wise(x_specs, self.top_k)
                    x_specs = x_specs * x_specs_mask
                x = x_nn * x_specs_mask
                x = self.dropout(x)
            x = self.layers[-1](x)
            x = self.dropout(x) 
            return x


def MLPMixer(*, image_size, channels, patch_size, dim, depth, num_classes, expansion_factor = 4, expansion_factor_token = 0.5, dropout = 0.):
    image_h, image_w = pair(image_size)
    assert (image_h % patch_size) == 0 and (image_w % patch_size) == 0, 'image must be divisible by patch size'
    num_patches = (image_h // patch_size) * (image_w // patch_size)
    chan_first, chan_last = partial(nn.Conv1d, kernel_size = 1), nn.Linear

    if specificity == True:
        feed_forward_instances = [FeedForward(dim, expansion_factor_token, dropout) for _ in range(depth)] # for specificity

        return nn.Sequential(
            Rearrange('b c (h p1) (w p2) -> b (h w) (p1 p2 c)', p1 = patch_size, p2 = patch_size),
            nn.Linear((patch_size ** 2) * channels, dim),
            *[nn.Sequential(
                # PreNormResidual(dim, FeedForward(num_patches, expansion_factor, dropout, chan_first)),
                PreNormResidual(dim, OrigFeedForward(num_patches, expansion_factor, dropout, chan_first)), # this uses conv1D, so we need to keep the GELU and stuff
                # PreNormResidual(dim, FeedForward(dim, expansion_factor_token, dropout, chan_last)) # standard
                PreNormResidual(dim, feed_forward_instances[i]) # this is what I will change with the standard vs spec
            ) for i in range(depth)],
            nn.LayerNorm(dim),
            Reduce('b n c -> b c', 'mean'),
            nn.Linear(dim, num_classes)
        )

    else: # standard
        return nn.Sequential(
            Rearrange('b c (h p1) (w p2) -> b (h w) (p1 p2 c)', p1 = patch_size, p2 = patch_size),
            nn.Linear((patch_size ** 2) * channels, dim),
            *[nn.Sequential(
                # PreNormResidual(dim, FeedForward(num_patches, expansion_factor, dropout, chan_first)),
                PreNormResidual(dim, OrigFeedForward(num_patches, expansion_factor, dropout, chan_first)), # this uses conv1D, so we need to keep the GELU and stuff
                PreNormResidual(dim, FeedForward(dim, expansion_factor_token, dropout, chan_last)) # standard
            ) for i in range(depth)],
            nn.LayerNorm(dim),
            Reduce('b n c -> b c', 'mean'),
            nn.Linear(dim, num_classes)
        )
        
